@@ -10,7 +10,7 @@ BEGIN {
 
 BEGIN {
 	$Type::Params::AUTHORITY = 'cpan:TOBYINK';
-	$Type::Params::VERSION   = '1.000005';
+	$Type::Params::VERSION   = '1.002001';
 }
 
 use B qw();
@@ -19,23 +19,40 @@ use Scalar::Util qw(refaddr);
 use Error::TypeTiny;
 use Error::TypeTiny::Assertion;
 use Error::TypeTiny::WrongNumberOfParameters;
-use Type::Tiny::Union;
 use Types::Standard -types;
 use Types::TypeTiny qw(CodeLike ArrayLike to_TypeTiny);
 
 require Exporter::Tiny;
 our @ISA = 'Exporter::Tiny';
 
-our @EXPORT = qw( compile );
-our @EXPORT_OK = qw( multisig validate Invocant );
+our @EXPORT    = qw( compile compile_named );
+our @EXPORT_OK = qw( multisig validate validate_named Invocant );
 
-BEGIN {
-	my $Invocant = 'Type::Tiny::Union'->new(
-		name             => 'Invocant',
-		type_constraints => [Object, ClassName],
-	);
-	sub Invocant () { $Invocant };
-};
+sub english_list {
+	require Type::Utils;
+	goto \&Type::Utils::english_list;
+}
+
+my $QUOTE = ($^V < 5.010 && exists(&B::cstring))
+	? \&B::cstring
+	: \&B::perlstring;   # is buggy on Perl 5.8
+
+{
+	my $Invocant;
+	sub Invocant () {
+		$Invocant ||= do {
+			require Type::Tiny::Union;
+			require Types::Standard;
+			'Type::Tiny::Union'->new(
+				name             => 'Invocant',
+				type_constraints => [
+					Types::Standard::Object(),
+					Types::Standard::ClassName(),
+				],
+			);
+		};
+	}
+}
 
 sub _mkslurpy
 {
@@ -52,7 +69,7 @@ sub _mkslurpy
 			'%s = (($#_-%d)%%2)==0 ? "Error::TypeTiny::WrongNumberOfParameters"->throw(message => sprintf("Odd number of elements in %%s", %s)) : +{ @_[%d..$#_] };',
 			$name,
 			$i,
-			B::perlstring("$tc"),
+			$QUOTE->("$tc"),
 			$i,
 		);
 }
@@ -60,8 +77,7 @@ sub _mkslurpy
 sub compile
 {
 	my (@code, %env);
-	@code = 'my (@R, %tmp, $tmp);';
-	push @code, '#placeholder';   # $code[1]
+	push @code, '#placeholder', '#placeholder';  # @code[0,1]
 	
 	my %options    = (ref($_[0]) eq "HASH" && !$_[0]{slurpy}) ? %{+shift} : ();
 	my $arg        = -1;
@@ -70,6 +86,23 @@ sub compile
 	my $max_args   = 0;
 	my $saw_opt    = 0;
 	
+	my $return_default_list = !!1;
+	$code[0] = 'my (%tmp, $tmp);';
+	PARAM: for my $param (@_) {
+		if (HashRef->check($param)) {
+			$code[0] = 'my (@R, %tmp, $tmp);';
+			$return_default_list = !!0;
+			last PARAM;
+		}
+		elsif (not Bool->check($param)) {
+			if ($param->has_coercion) {
+				$code[0] = 'my (@R, %tmp, $tmp);';
+				$return_default_list = !!0;
+				last PARAM;
+			}
+		}
+	}
+		
 	while (@_)
 	{
 		++$arg;
@@ -86,7 +119,10 @@ sub compile
 		
 		if (HashRef->check($constraint))
 		{
-			$constraint = to_TypeTiny($constraint->{slurpy});
+			$constraint = to_TypeTiny(
+				$constraint->{slurpy}
+					or Error::TypeTiny::croak("Slurpy parameter malformed")
+			);
 			push @code,
 				$constraint->is_a_type_of(Dict)     ? _mkslurpy('$_', '%', $constraint => $arg) :
 				$constraint->is_a_type_of(Map)      ? _mkslurpy('$_', '%', $constraint => $arg) :
@@ -107,7 +143,11 @@ sub compile
 			
 			if ($is_optional)
 			{
-				push @code, sprintf 'return @R if $#_ < %d;', $arg;
+				push @code, sprintf(
+					'return %s if $#_ < %d;',
+					$return_default_list ? '@_' : '@R',
+					$arg,
+				);
 				$saw_opt++;
 				$max_args++;
 			}
@@ -150,7 +190,7 @@ sub compile
 					? $constraint->type_parameter->inline_check($varname)
 					: $constraint->inline_check($varname),
 				$constraint->{uniq},
-				B::perlstring($constraint),
+				$QUOTE->($constraint),
 				$varname,
 				$is_slurpy ? 'q{$SLURPY}' : sprintf('q{$_[%d]}', $arg),
 			);
@@ -164,13 +204,15 @@ sub compile
 				'%s or Type::Tiny::_failed_check(%d, %s, %s, varname => %s);',
 				sprintf(sprintf '$check[%d]->(%s)', $arg, $varname),
 				$constraint->{uniq},
-				B::perlstring($constraint),
+				$QUOTE->($constraint),
 				$varname,
 				$is_slurpy ? 'q{$SLURPY}' : sprintf('q{$_[%d]}', $arg),
 			);
 		}
 		
-		push @code, sprintf 'push @R, %s;', $varname;
+		unless ($return_default_list) {
+			push @code, sprintf 'push @R, %s;', $varname;
+		}
 	}
 	
 	if ($min_args == $max_args and not $saw_slurpy)
@@ -201,7 +243,12 @@ sub compile
 		);
 	}
 	
-	push @code, '@R;';
+	if ($return_default_list) {
+		push @code, '@_;';
+	}
+	else {
+		push @code, '@R;';
+	}
 	
 	my $source  = "sub { no warnings; ".join("\n", @code)." };";
 	
@@ -222,11 +269,162 @@ sub compile
 	return $closure;
 }
 
+sub compile_named
+{
+	my (@code, %env);
+	
+	@code = 'my (%R, %tmp, $tmp);';
+	push @code, '#placeholder';   # $code[1]
+	
+	my %options    = (ref($_[0]) eq "HASH" && !$_[0]{slurpy}) ? %{+shift} : ();
+	my $arg = -1;
+	my $had_slurpy;
+	
+	push @code, 'my %in = ((@_==1) && ref($_[0]) eq "HASH") ? %{$_[0]} : (@_ % 2) ? "Error::TypeTiny::WrongNumberOfParameters"->throw(message => "Odd number of elements in hash") : @_;';
+	
+	while (@_) {
+		++$arg;
+		my ($name, $constraint) = splice(@_, 0, 2);
+		
+		my $is_optional;
+		my $really_optional;
+		my $is_slurpy;
+		my $varname;
+		
+		if (Bool->check($constraint))
+		{
+			$constraint = $constraint ? Any : Optional[Any];
+		}
+		
+		if (HashRef->check($constraint))
+		{
+			$constraint = to_TypeTiny($constraint->{slurpy});
+			++$is_slurpy;
+			++$had_slurpy;
+		}
+		else
+		{
+			$is_optional     = grep $_->{uniq} == Optional->{uniq}, $constraint->parents;
+			$really_optional = $is_optional && $constraint->parent->{uniq} eq Optional->{uniq} && $constraint->type_parameter;
+			
+			$constraint = $constraint->type_parameter if $really_optional;
+		}
+		
+		unless ($is_optional or $is_slurpy) {
+			push @code, sprintf(
+				'exists($in{%s}) or "Error::TypeTiny::WrongNumberOfParameters"->throw(message => sprintf "Missing required parameter: %%s", %s);',
+				$QUOTE->($name),
+				$QUOTE->($name),
+			);
+		}
+		
+		my $need_to_close_if = 0;
+		
+		if ($is_slurpy) {
+			$varname = '\\%in';
+		}
+		elsif ($is_optional) {
+			push @code, sprintf('if (exists($in{%s})) {', $QUOTE->($name));
+			push @code, sprintf('$tmp = delete($in{%s});', $QUOTE->($name));
+			$varname = '$tmp';
+			++$need_to_close_if;
+		}
+		else {
+			push @code, sprintf('$tmp = delete($in{%s});', $QUOTE->($name));
+			$varname = '$tmp';
+		}
+		
+		if ($constraint->has_coercion) {
+			if ($constraint->coercion->can_be_inlined) {
+				push @code, sprintf(
+					'$tmp = %s;',
+					$constraint->coercion->inline_coercion($varname)
+				);
+			}
+			else {
+				$env{'@coerce'}[$arg] = $constraint->coercion->compiled_coercion;
+				push @code, sprintf(
+					'$tmp = $coerce[%d]->(%s);',
+					$arg,
+					$varname,
+				);
+			}
+			$varname = '$tmp';
+		}
+		
+		if ($constraint->can_be_inlined)
+		{
+			push @code, sprintf(
+				'(%s) or Type::Tiny::_failed_check(%d, %s, %s, varname => %s);',
+				$constraint->inline_check($varname),
+				$constraint->{uniq},
+				$QUOTE->($constraint),
+				$varname,
+				$is_slurpy ? 'q{$SLURPY}' : sprintf('q{$_{%s}}', $QUOTE->($name)),
+			);
+		}
+		else
+		{
+			$env{'@check'}[$arg] = $constraint->compiled_check;
+			push @code, sprintf(
+				'%s or Type::Tiny::_failed_check(%d, %s, %s, varname => %s);',
+				sprintf(sprintf '$check[%d]->(%s)', $arg, $varname),
+				$constraint->{uniq},
+				$QUOTE->($constraint),
+				$varname,
+				$is_slurpy ? 'q{$SLURPY}' : sprintf('q{$_{%s}}', $QUOTE->($name)),
+			);
+		}
+		
+		push @code, sprintf('$R{%s} = %s;', $QUOTE->($name), $varname);
+		
+		push @code, '}' if $need_to_close_if;
+	}
+	
+	if (!$had_slurpy) {
+		push @code, 'keys(%in) and "Error::TypeTiny"->throw(message => sprintf "Unrecognized parameter%s: %s", keys(%in)>1?"s":"", Type::Params::english_list(sort keys %in));'
+	}
+	
+	push @code, '\\%R;';
+	
+	my $source  = "sub { no warnings; ".join("\n", @code)." };";
+	return $source if $options{want_source};
+	
+	my $closure = eval_closure(
+		source      => $source,
+		description => sprintf("parameter validation for '%s'", [caller(1+($options{caller_level}||0))]->[3] || '__ANON__'),
+		environment => \%env,
+	);
+	
+	return {
+		min_args   => undef,  # always going to be 1 or 0
+		max_args   => undef,  # should be possible to figure out if no slurpy param
+		closure    => $closure,
+	} if $options{want_details};
+	
+	return $closure;
+}
+
 my %compiled;
 sub validate
 {
 	my $arr = shift;
-	my $sub = $compiled{ join ":", map($_->{uniq}||"\@$_->{slurpy}", @_) } ||= compile({ caller_level => 1 }, @_);
+	my $sub = (
+		$compiled{ join ":", map($_->{uniq}||"\@$_->{slurpy}", @_) }
+			||= compile({ caller_level => 1 }, @_)
+	);
+	@_ = @$arr;
+	goto $sub;
+}
+
+my %compiled_named;
+sub validate_named
+{
+	my $arr = shift;
+	my $sub = (
+		$compiled_named{ join ":", map(ref($_)?($_->{uniq}||"\@$_->{slurpy}"):$QUOTE->($_), @_) }
+			||= compile_named({ caller_level => 1 }, @_)
+	);
 	@_ = @$arr;
 	goto $sub;
 }
@@ -244,12 +442,17 @@ sub multisig
 	
 	for my $i (0 .. $#multi)
 	{
-		my $sig = $multi[$i];
+		my $flag = sprintf('${^TYPE_PARAMS_MULTISIG} = %d', $i);
+		my $sig  = $multi[$i];
 		my @cond;
 		push @cond, sprintf('@_ >= %s', $sig->{min_args}) if defined $sig->{min_args};
 		push @cond, sprintf('@_ <= %s', $sig->{max_args}) if defined $sig->{max_args};
+		if (defined $sig->{max_args} and defined $sig->{min_args}) {
+			@cond = sprintf('@_ == %s', $sig->{min_args})
+				if $sig->{max_args} == $sig->{min_args};
+		}
 		push @code, sprintf('if (%s){', join(' and ', @cond)) if @cond;
-		push @code, sprintf('eval { $r = [ $multi[%d]{closure}->(@_) ] };', $i);
+		push @code, sprintf('eval { $r = [ $multi[%d]{closure}->(@_) ]; %s };', $i, $flag);
 		push @code, 'return(@$r) if $r;';
 		push @code, '}' if @cond;
 	}
@@ -272,7 +475,7 @@ __END__
 
 =encoding utf-8
 
-=for stopwords evals
+=for stopwords evals invocant
 
 =head1 NAME
 
@@ -334,7 +537,7 @@ example from the SYNOPSIS could be rewritten as:
  sub deposit_monies
  {
     $deposit_monies_check ||= compile( Str, Str, slurpy ArrayRef[Num] );
-    my ($sort_code, $account_number, $monies) = $check->(@_);
+    my ($sort_code, $account_number, $monies) = $deposit_monies_check->(@_);
     
     ...;
  }
@@ -343,7 +546,7 @@ Not quite as neat, but not awful either.
 
 There's a shortcut reducing it to one step:
 
- use Type::Params qw( validate );
+ use Type::Params qw( validate validate_named );
  
  sub deposit_monies
  {
@@ -365,6 +568,10 @@ Dude, these functions are documented!
 
 =item validate
 
+=item compile_named
+
+=item validate_named
+
 =item Invocant
 
 =item multisig
@@ -420,6 +627,19 @@ a type constraint accepting blessed objects and also class names.
       $self->_run;
    }
 
+Of course, some people like to use C<shift> for the invocant:
+
+   sub dump
+   {
+      my $self = shift;
+      
+      state $check = compile( Int );
+      my ($limit) = $check->(@_);
+      
+      local $Data::Dumper::Maxdepth = $limit;
+      print Data::Dumper::Dumper($self);
+   }
+
 =head2 Optional Parameters
 
    use Types::Standard qw( Object Optional Int );
@@ -458,10 +678,38 @@ will die if an odd number of elements are slurped in.
 A check may only have one slurpy parameter, and it must be the last
 parameter.
 
+Having a slurpy parameter will slightly slow down your checks.
+
 =head2 Named Parameters
 
-Just use a slurpy C<Dict>:
+You can use C<compile_named> to accept a hash of named parameters
 
+   use Type::Params qw(compile_named);
+   use Types::Standard qw( slurpy Dict Ref Optional Int );
+   
+   sub dump
+   {
+      state $check = compile_named(
+         var    => Ref,
+         limit  => Optional[Int],
+      );
+      my $arg = $check->(@_);
+      
+      local $Data::Dumper::Maxdepth = $arg->{limit};
+      print Data::Dumper::Dumper($arg->{var});
+   }
+   
+   dump({ var => $foo, limit => 1 });    # ok (hashref)
+   dump(  var => $foo, limit => 1  );    # ok (hash)
+   dump(  var => $foo  );                # ok (no optional parameter)
+   dump(  limit => 1  );                 # dies
+
+Prior to Type::Tiny 1.002000, the recommendation was to use a slurpy
+C<Dict>. This still works, though the error messages you get might not
+be quite so nice, and you don't get the automatic detection of hash
+versus hashref in the input C<< @_ >>. Oh, and it's usually slower.
+
+   use Type::Params qw(compile);
    use Types::Standard qw( slurpy Dict Ref Optional Int );
    
    sub dump
@@ -478,12 +726,14 @@ Just use a slurpy C<Dict>:
       print Data::Dumper::Dumper($arg->{var});
    }
    
-   dump(var => $foo, limit => 1);   # ok
-   dump(var => $foo);               # ok
-   dump(limit => 1);                # dies
+   dump(  var => $foo, limit => 1  );    # ok (hash)
+   dump(  var => $foo  );                # ok (no optional parameter)
+   dump(  limit => 1  );                 # dies
 
 =head2 Mixed Positional and Named Parameters
 
+For this, you can still use the C<< slurpy Dict >> hack...
+
    use Types::Standard qw( slurpy Dict Ref Optional Int );
    
    sub my_print
@@ -496,6 +746,8 @@ Just use a slurpy C<Dict>:
          ],
       );
       my ($string, $arg) = $check->(@_);
+      
+      ...;
    }
    
    my_print("Hello World", colour => "blue");
@@ -553,6 +805,8 @@ Note that the coercion is specified as a string of Perl code. This is usually
 the fastest way to do it, but a coderef is also accepted. Either way, the
 value to be coerced is C<< $_ >>.
 
+Having any coercions will slightly slow down your checks.
+
 =head2 Alternatives
 
 Type::Params can export a C<multisig> function that compiles multiple
@@ -572,8 +826,8 @@ alternative signatures into one, and uses the first one that works:
 
 Coercions, slurpy parameters, etc still work.
 
-There's currently no indication of which of the multiple signatures
-succeeded.
+The magic global C<< ${^TYPE_PARAMS_MULTISIG} >> is set to the index of
+the first signature which succeeded.
 
 The present implementation involves compiling each signature independently,
 and trying them each (in their given order!) in an C<eval> block. The only
@@ -588,6 +842,7 @@ It's also possible to list coderefs as alternatives in C<multisig>:
       sub { ... },
       [ HashRef, Num ],
       [ CodeRef ],
+      compile_named( needle => Value, haystack => Ref ),
    );
 
 The coderef is expected to die if that alternative should be abandoned (and
@@ -605,18 +860,46 @@ a full example:
             return ($meth, $obj);
          },
       );
+      
       my ($needle, $haystack) = $check->(@_);
       
-      is_HashRef($haystack)  ? $haystack->{$needle} :
-      is_ArrayRef($haystack) ? $haystack->[$needle] :
-      is_Object($haystack)   ? $haystack->$needle   :
-      die;
+      for (${^TYPE_PARAMS_MULTISIG) {
+         return $haystack->[$needle] if $_ == 0;
+         return $haystack->{$needle} if $_ == 1;
+         return $haystack->$needle   if $_ == 2;
+      }
    }
    
    get_from(0, \@array);      # returns $array[0]
    get_from('foo', \%hash);   # returns $hash{foo}
    get_from('foo', $obj);     # returns $obj->foo
 
+=head2 Defaults
+
+Type::Params does not currently offer a built-in way to set defaults
+for a parameter. Setting defaults manually is not especially difficult.
+
+   sub print_coloured {
+      state $check = compile( Str, Optional[Str] );
+      
+      my ($text, $colour) = $check->(@_);
+      $colour //= "black";
+      
+      ...;
+   }
+
+I occasionally get requests for this to work:
+
+   sub print_coloured {
+      state $check = compile( Str, Default[Str, "black"] );
+      
+      my ($text, $colour) = $check->(@_);
+      
+      ...;
+   }
+
+But honestly, I don't find that any clearer.
+
 =head1 COMPARISON WITH PARAMS::VALIDATE
 
 L<Type::Params> is not really a drop-in replacement for L<Params::Validate>;
@@ -634,11 +917,8 @@ first time your sub is called to make subsequent calls a lot faster.
 
 =item *
 
-Type::Params is mostly geared towards positional parameters, while
-Params::Validate seems to be primarily aimed at named parameters. (Though
-either works for either.) Params::Validate doesn't appear to have a
-particularly natural way of validating a mix of positional and named
-parameters.
+Params::Validate doesn't appear to have a particularly natural way of
+validating a mix of positional and named parameters.
 
 =item *
 
@@ -671,6 +951,47 @@ any other Type::Tiny type constraints.
 
 =back
 
+=head1 COMPARISON WITH PARAMS::VALIDATIONCOMPILER
+
+L<Params::ValidationCompiler> does basically the same thing as
+L<Type::Params>.
+
+=over
+
+=item *
+
+Params::ValidationCompiler and Type::Params are likely to perform fairly
+similarly. In most cases, recent versions of Type::Params seem to be
+I<slightly> faster, but except in very trivial cases, you're unlikely to
+notice the speed difference. Speed probably shouldn't be a factor when
+choosing between them.
+
+=item *
+
+Type::Params's syntax is more compact:
+
+   state $check = compile(Object, Optional[Int], slurpy ArrayRef);
+
+Versus:
+
+   state $check = validation_for(
+      params => [
+         { type => Object },
+         { type => Int,      optional => 1 },
+         { type => ArrayRef, slurpy => 1 },
+      ],
+   );
+
+=item *
+
+L<Params::ValidationCompiler> offers defaults.
+
+=item *
+
+L<Params::ValidationCompiler> probably has slightly better exceptions.
+
+=back
+
 =head1 BUGS
 
 Please report any bugs to
@@ -686,7 +1007,7 @@ Toby Inkster E<lt>tobyink@cpan.orgE<gt>.
 
 =head1 COPYRIGHT AND LICENCE
 
-This software is copyright (c) 2013-2014 by Toby Inkster.
+This software is copyright (c) 2013-2014, 2017 by Toby Inkster.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
